from __future__ import annotations

from collections import Counter
from functools import reduce

import numpy as np
import torch
from torch import Tensor

from qadence import BackendName, DiffMode
from qadence.backends import backend_factory
from qadence.blocks.abstract import AbstractBlock
from qadence.blocks.block_to_tensor import (
    HMAT,
    IMAT,
    SDAGMAT,
    ZMAT,
    block_to_tensor,
)
from qadence.blocks.composite import CompositeBlock
from qadence.blocks.primitive import PrimitiveBlock
from qadence.blocks.utils import get_pauli_blocks, unroll_block_with_scaling
from qadence.circuit import QuantumCircuit
from qadence.operations import X, Y, Z, chain, kron
from qadence.states import one_state, zero_state
from qadence.types import Endianness

pauli_gates = [X, Y, Z]


UNITARY_TENSOR = [
    ZMAT @ HMAT,
    SDAGMAT.squeeze(dim=0) @ HMAT,
    IMAT,
]


# Projector matrices in Big-Endian convention.
PROJECTOR_MATRICES = {
    "0": zero_state(n_qubits=1).t() @ zero_state(n_qubits=1),
    "1": one_state(n_qubits=1).t() @ one_state(n_qubits=1),
}


def identity(n_qubits: int) -> Tensor:
    return torch.eye(2**n_qubits, dtype=torch.complex128)


def _max_observable_weight(observable: AbstractBlock) -> int:
    """
    Get the maximal weight for the given observable.

    The weight is a measure of the locality of the observable,
    a count of the number of qubits on which the observable acts
    non-trivially.

    See https://arxiv.org/pdf/2002.08953.pdf
    Supplementary Material 1 and Eq. (S17).
    """
    pauli_decomposition = unroll_block_with_scaling(observable)
    weights = []
    for pauli_term in pauli_decomposition:
        weight = 0
        block = pauli_term[0]
        if isinstance(block, PrimitiveBlock):
            if isinstance(block, (X, Y, Z)):
                weight += 1
            weights.append(weight)
        else:
            pauli_blocks = get_pauli_blocks(block=block)
            weight = 0
            for block in pauli_blocks:
                if isinstance(block, (X, Y, Z)):
                    weight += 1
            weights.append(weight)
    return max(weights)


def maximal_weight(observables: list[AbstractBlock]) -> int:
    """Return the maximal weight if a list of observables is provided."""
    return max([_max_observable_weight(observable=observable) for observable in observables])


def number_of_samples(
    observables: list[AbstractBlock], accuracy: float, confidence: float
) -> tuple[int, ...]:
    """
    Estimate an optimal shot budget and a shadow partition size
    to guarantee given accuracy on all observables expectation values
    within 1 - confidence range.

    See https://arxiv.org/pdf/2002.08953.pdf
    Supplementary Material 1 and Eqs. (S23)-(S24).
    """
    max_k = maximal_weight(observables=observables)
    N = round(3**max_k * 34.0 / accuracy**2)
    K = round(2.0 * np.log(2.0 * len(observables) / confidence))
    return N, K


def local_shadow(sample: Counter, unitary_ids: list) -> Tensor:
    """
    Compute local shadow by inverting the quantum channel for each projector state.

    See https://arxiv.org/pdf/2002.08953.pdf
    Supplementary Material 1 and Eqs. (S17,S44).

    Expects a sample bitstring in ILO.
    """
    bitstring = list(sample.keys())[0]
    local_density_matrices = []
    for bit, unitary_id in zip(bitstring, unitary_ids):
        proj_mat = PROJECTOR_MATRICES[bit]
        unitary_tensor = UNITARY_TENSOR[unitary_id].squeeze(dim=0)
        local_density_matrices.append(
            3 * (unitary_tensor.adjoint() @ proj_mat @ unitary_tensor) - identity(1)
        )
    if len(local_density_matrices) == 1:
        return local_density_matrices[0]
    else:
        return reduce(torch.kron, local_density_matrices)


def classical_shadow(
    shadow_size: int,
    circuit: QuantumCircuit,
    param_values: dict,
    state: Tensor | None = None,
    backend_name: BackendName = BackendName.PYQTORCH,
    # FIXME: Changed below from Little to Big, double-check when Roland is back
    endianness: Endianness = Endianness.BIG,
) -> list:
    shadow: list = []
    backend = backend_factory(backend=backend_name, diff_mode=DiffMode.GPSR)
    # TODO: Parallelize embarrassingly parallel loop.
    for _ in range(shadow_size):
        unitary_ids = np.random.randint(0, 3, size=(1, circuit.n_qubits))[0]
        random_unitary = [
            pauli_gates[unitary_ids[qubit]](qubit) for qubit in range(circuit.n_qubits)
        ]
        if len(random_unitary) == 1:
            random_unitary_block = random_unitary[0]
        else:
            random_unitary_block = kron(*random_unitary)
        rotated_circuit = QuantumCircuit(
            circuit.n_qubits,
            chain(circuit.block, random_unitary_block),
        )
        # Reverse endianness to get sample bitstrings in ILO.
        conv_circ = backend.circuit(rotated_circuit)
        samples = backend.sample(
            circuit=conv_circ,
            param_values=param_values,
            n_shots=1,
            state=state,
            endianness=endianness,
        )
        batched_shadow = []
        for batch in samples:
            batched_shadow.append(local_shadow(sample=batch, unitary_ids=unitary_ids))
        shadow.append(batched_shadow)

    # Reshape the shadow by batches of samples instead of samples of batches.
    # FIXME: Improve performance.
    return [list(s) for s in zip(*shadow)]


def reconstruct_state(shadow: list) -> Tensor:
    """Reconstruct the state density matrix for the given shadow."""
    return reduce(torch.add, shadow) / len(shadow)


def compute_traces(
    qubit_support: tuple,
    N: int,
    K: int,
    shadow: list,
    observable: AbstractBlock,
    endianness: Endianness = Endianness.BIG,
) -> list:
    floor = int(np.floor(N / K))
    traces = []
    # TODO: Parallelize embarrassingly parallel loop.
    for k in range(K):
        reconstructed_state = reconstruct_state(shadow=shadow[k * floor : (k + 1) * floor])
        # Reshape the observable matrix to fit the density matrix dimensions
        # by filling indentites.
        # Please note the endianness is also flipped to get results in LE.
        # FIXME: Changed below from Little to Big, double-check when Roland is back
        # FIXME: Correct these comments.
        trace = (
            (
                block_to_tensor(
                    block=observable,
                    qubit_support=qubit_support,
                    endianness=Endianness.BIG,
                ).squeeze(dim=0)
                @ reconstructed_state
            )
            .trace()
            .real
        )
        traces.append(trace)
    return traces


def estimators(
    qubit_support: tuple,
    N: int,
    K: int,
    shadow: list,
    observable: AbstractBlock,
    endianness: Endianness = Endianness.BIG,
) -> Tensor:
    """
    Return estimators (traces of observable times mean density matrix)
    for K equally-sized shadow partitions.

    See https://arxiv.org/pdf/2002.08953.pdf
    Algorithm 1.
    """
    # If there is no Pauli-Z operator in the observable,
    # the sample can't "hit" that measurement.
    if isinstance(observable, PrimitiveBlock):
        if type(observable) == Z:
            traces = compute_traces(
                qubit_support=qubit_support,
                N=N,
                K=K,
                shadow=shadow,
                observable=observable,
                endianness=endianness,
            )
        else:
            traces = [torch.tensor(0.0)]
    elif isinstance(observable, CompositeBlock):
        if Z in observable:
            traces = compute_traces(
                qubit_support=qubit_support,
                N=N,
                K=K,
                shadow=shadow,
                observable=observable,
                endianness=endianness,
            )
        else:
            traces = [torch.tensor(0.0)]
    return torch.tensor(traces, dtype=torch.get_default_dtype())


def estimations(
    circuit: QuantumCircuit,
    observables: list[AbstractBlock],
    param_values: dict,
    shadow_size: int | None = None,
    accuracy: float = 0.1,
    confidence: float = 0.1,
    state: Tensor | None = None,
    backend_name: BackendName = BackendName.PYQTORCH,
    endianness: Endianness = Endianness.BIG,
) -> Tensor:
    """Compute expectation values for all local observables using median of means."""
    # N is the estimated shot budget for the classical shadow to
    # achieve desired accuracy for all L = len(observables) within 1 - confidence probablity.
    # K is the size of the shadow partition.
    N, K = number_of_samples(observables=observables, accuracy=accuracy, confidence=confidence)
    if shadow_size is not None:
        N = shadow_size
    shadow = classical_shadow(
        shadow_size=N,
        circuit=circuit,
        param_values=param_values,
        state=state,
        backend_name=backend_name,
        endianness=endianness,
    )
    estimations = []
    for observable in observables:
        pauli_decomposition = unroll_block_with_scaling(observable)
        batch_estimations = []
        for batch in shadow:
            pauli_term_estimations = []
            for pauli_term in pauli_decomposition:
                # Get the estimators for the current Pauli term.
                # This is a tensor<float> of size K.
                estimation = estimators(
                    qubit_support=circuit.block.qubit_support,
                    N=N,
                    K=K,
                    shadow=batch,
                    observable=pauli_term[0],
                    endianness=endianness,
                )
                # Compute the median of means for the current Pauli term.
                # Weigh the median by the Pauli term scaling.
                pauli_term_estimations.append(torch.median(estimation) * pauli_term[1])
            # Sum the expectations for each Pauli term to get the expectation for the
            # current batch.
            batch_estimations.append(sum(pauli_term_estimations))
        estimations.append(batch_estimations)
    return torch.transpose(torch.tensor(estimations, dtype=torch.get_default_dtype()), 1, 0)


def compute_expectation(
    circuit: QuantumCircuit,
    observables: list[AbstractBlock],
    param_values: dict,
    options: dict,
    state: Tensor | None = None,
    backend_name: BackendName = BackendName.PYQTORCH,
    endianness: Endianness = Endianness.BIG,
) -> Tensor:
    """
    Construct a classical shadow of a state to estimate observable expectation values.

    Args:
        circuit (QuantumCircuit): a circuit to prepare the state.
        observables (list[AbstractBlock]): a list of observables
            to estimate the expectation values from.
        param_values (dict): a dict of values to substitute the
            symbolic parameters for.
        options (dict): a dict of options for the measurement protocol.
            Here, shadow_size (int), accuracy (float) and confidence (float) are supported.
        state (Tensor | None): an initial input state.
        backend_name (BackendName): a backend name to retrieve computations from.

    Returns:
        expectations (Tensor): an estimation of the expectation values.
    """
    if not isinstance(observables, list):
        raise TypeError(
            "Observables must be of type <class 'List[AbstractBlock]'>. Got {}.".format(
                type(observables)
            )
        )
    shadow_size = options.get("shadow_size", None)
    accuracy = options.get("accuracy", None)
    if shadow_size is None and accuracy is None:
        KeyError(
            "Shadow protocol requires either an option"
            "'shadow_size' of type 'int' or 'accuracy' of type 'float'."
        )
    confidence = options.get("confidence", None)
    if confidence is None:
        KeyError("Shadow protocol requires a 'confidence' kwarg of type 'float'.")
    return estimations(
        circuit=circuit,
        observables=observables,
        param_values=param_values,
        shadow_size=shadow_size,
        accuracy=accuracy,
        confidence=confidence,
        state=state,
        backend_name=backend_name,
        endianness=endianness,
    )



from __future__ import annotations

from typing import Callable

from torch import Tensor

from qadence.backend import BackendConfiguration, BackendName
from qadence.backends.pytorch_wrapper import DiffMode
from qadence.blocks import AbstractBlock
from qadence.circuit import QuantumCircuit
from qadence.measurements import Measurements
from qadence.models.quantum_model import QuantumModel
from qadence.utils import Endianness


class QNN(QuantumModel):
    """Quantum neural network model for n-dimensional inputs

    Examples:
    ```python exec="on" source="material-block" result="json"
    import torch
    from qadence import QuantumCircuit, QNN
    from qadence import hea, feature_map, hamiltonian_factory, Z

    # create the circuit
    n_qubits, depth = 2, 4
    fm = feature_map(n_qubits)
    ansatz = hea(n_qubits=n_qubits, depth=depth)
    circuit = QuantumCircuit(n_qubits, fm, ansatz)
    obs_base = hamiltonian_factory(n_qubits, detuning = Z)

    # the QNN will yield two outputs
    obs = [2.0 * obs_base, 4.0 * obs_base]

    # initialize and use the model
    qnn = QNN(circuit, obs, diff_mode="ad", backend="pyqtorch")
    y = qnn.expectation({"phi": torch.rand(3)})
    print(str(y)) # markdown-exec: hide
    ```
    """

    def __init__(
        self,
        circuit: QuantumCircuit,
        observable: list[AbstractBlock] | AbstractBlock,
        transform: Callable[[Tensor], Tensor] = None,  # transform output of the QNN
        backend: BackendName = BackendName.PYQTORCH,
        diff_mode: DiffMode = DiffMode.AD,
        protocol: Measurements | None = None,
        configuration: BackendConfiguration | dict | None = None,
    ):
        """Initialize the QNN

        The number of inputs is determined by the feature parameters in the input
        quantum circuit while the number of outputs is determined by how many
        observables are provided as input

        Args:
            circuit: The quantum circuit to use for the QNN.
            transform: A transformation applied to the output of the QNN.
            backend: The chosen quantum backend.
            diff_mode: The differentiation engine to use. Choices 'gpsr' or 'ad'.
            protocol: optional measurement protocol. If None,
                use exact expectation value with a statevector simulator
            configuration: optional configuration for the backend

        """
        super().__init__(
            circuit=circuit,
            observable=observable,
            backend=backend,
            diff_mode=diff_mode,
            protocol=protocol,
            configuration=configuration,
        )

        if self.out_features is None:
            raise ValueError("You need to provide at least one observable in the QNN constructor")

        self.transform = transform if transform else lambda x: x

    def forward(
        self,
        values: dict[str, Tensor] | Tensor = None,
        state: Tensor | None = None,
        protocol: Measurements | None = None,
        endianness: Endianness = Endianness.BIG,
    ) -> Tensor:
        """Forward pass of the model

        This returns the (differentiable) expectation value of the given observable
        operator defined in the constructor. Differently from the base QuantumModel
        class, the QNN accepts also a tensor as input for the forward pass. The
        tensor is expected to have shape: `n_batches x in_features` where `n_batches`
        is the number of data points and `in_features` is the dimensionality of the problem

        The output of the forward pass is the expectation value of the input
        observable(s). If a single observable is given, the output shape is
        `n_batches` while if multiple observables are given the output shape
        is instead `n_batches x n_observables`

        Args:
            values (dict[str, Tensor] | Tensor): the values of the feature parameters

        Returns:
            Tensor: a tensor with the expectation value of the observables passed
                in the constructor of the model
        """
        if values is None:
            values = {}
        if not isinstance(values, dict):
            values = self._format_to_dict(values)
        if protocol is None:
            protocol = self._protocol

        return self.transform(
            self.expectation(values=values, state=state, protocol=protocol, endianness=endianness)
        )

    def _format_to_dict(self, values: Tensor) -> dict[str, Tensor]:
        """Format an input tensor into the format required by the forward pass

        The tensor is assumed to have dimensions: n_batches x in_features where in_features
        corresponds to the number of input features of the QNN
        """

        if len(values.size()) == 1:
            values = values.reshape(-1, 1)
        msg = f"Model expects in_features={self.in_features} but got {values.size()[1]}."
        assert len(values.size()) == 2, msg
        assert values.size()[1] == self.in_features, msg

        names = [p.name for p in self.inputs]
        res = {}
        for i, name in enumerate(names):
            res[name] = values[:, i]
        return res

    # TODO: Implement derivatives w.r.t. to inputs



# flake8: noqa

from .feature_maps import (
    feature_map,
    chebyshev_feature_map,
    fourier_feature_map,
    tower_feature_map,
    exp_fourier_feature_map,
)

from .ansatze import hea, build_qnn

from .daqc import daqc_transform

from .hamiltonians import (
    hamiltonian_factory,
    ising_hamiltonian,
    single_z,
    total_magnetization,
    zz_hamiltonian,
)

from .qft import qft

# Modules to be automatically added to the qadence namespace
__all__ = [
    "feature_map",
    "chebyshev_feature_map",
    "fourier_feature_map",
    "tower_feature_map",
    "exp_fourier_feature_map",
    "hea",
    "build_qnn",
    "hamiltonian_factory",
    "ising_hamiltonian",
    "single_z",
    "total_magnetization",
    "zz_hamiltonian",
    "qft",
    "daqc_transform",
]



from __future__ import annotations

import json
import os
from pathlib import Path
from typing import Any, get_args
from typing import Union as TypingUnion

import torch
from sympy import *
from sympy import Basic, Expr, srepr

from qadence import QuantumCircuit, operations
from qadence import blocks as qadenceblocks
from qadence.blocks import AbstractBlock
from qadence.blocks.utils import tag
from qadence.logger import get_logger
from qadence.ml_tools.models import TransformedModule
from qadence.models import QNN, QuantumModel
from qadence.parameters import Parameter
from qadence.register import Register
from qadence.types import SerializationFormat

# Modules to be automatically added to the qadence namespace
__all__ = ["deserialize", "load", "save", "serialize"]


logger = get_logger(__name__)


def file_extension(file: Path | str) -> str:
    FORMAT = ""
    if isinstance(file, str):
        _, extension = os.path.splitext(file)
        FORMAT = extension[1:].upper()
    elif isinstance(file, os.PathLike):
        _, extension = os.path.splitext(str(file))
        FORMAT = extension[1:].upper()
    return FORMAT


SUPPORTED_OBJECTS = [
    AbstractBlock,
    QuantumCircuit,
    QuantumModel,
    QNN,
    TransformedModule,
    Register,
    Basic,
    torch.nn.Module,
]
SUPPORTED_TYPES = TypingUnion[
    AbstractBlock,
    QuantumCircuit,
    QuantumModel,
    QNN,
    TransformedModule,
    Register,
    Basic,
    torch.nn.Module,
]


ALL_BLOCK_NAMES = [
    n for n in dir(qadenceblocks) if not (n.startswith("__") and n.endswith("__"))
] + [n for n in dir(operations) if not (n.startswith("__") and n.endswith("__"))]


def save_pt(d: dict, file_path: str | Path) -> None:
    torch.save(d, file_path)


def save_json(d: dict, file_path: str | Path) -> None:
    with open(file_path, "w") as file:
        file.write(json.dumps(d))


def load_pt(file_path: str | Path, map_location: str) -> Any:
    return torch.load(file_path, map_location=map_location)


def load_json(file_path: str | Path, map_location: str) -> Any:
    with open(file_path, "r") as file:
        return json.load(file)


FORMAT_DICT = {
    SerializationFormat.PT: (".pt", save_pt, load_pt, True),
    SerializationFormat.JSON: (".json", save_json, load_json, False),
}


def serialize(obj: SUPPORTED_TYPES, save_params: bool = False) -> dict:
    """
    Supported Types:
    AbstractBlock | QuantumCircuit | QuantumModel | TransformedModule | Register | Module
    Serializes a qadence object to a dictionary.

    Arguments:
        obj (AbstractBlock | QuantumCircuit | QuantumModel | Register | Module):
    Returns:
        A dict.

    Examples:
    ```python exec="on" source="material-block" result="json"
    import torch
    from qadence import serialize, deserialize, hea, hamiltonian_factory, Z
    from qadence import QuantumCircuit, QuantumModel

    n_qubits = 2
    myblock = hea(n_qubits=n_qubits, depth=1)
    block_dict = serialize(myblock)
    print(block_dict)

    ## Lets use myblock in a QuantumCircuit and serialize it.

    qc = QuantumCircuit(n_qubits, myblock)
    qc_dict = serialize(qc)
    qc_deserialized = deserialize(qc_dict)
    assert qc == qc_deserialized

    ## Finally, let's wrap it in a QuantumModel
    obs = hamiltonian_factory(n_qubits, detuning = Z)
    qm = QuantumModel(qc, obs, backend='pyqtorch', diff_mode='ad')

    qm_dict = serialize(qm)
    qm_deserialized = deserialize(qm_dict)
    # Lets check if the loaded QuantumModel returns the same expectation
    assert torch.isclose(qm.expectation({}), qm_deserialized.expectation({}))
    ```
    """
    if not isinstance(obj, get_args(SUPPORTED_TYPES)):
        logger.error(TypeError(f"Serialization of object type {type(obj)} not supported."))
    d: dict = {}
    try:
        if isinstance(obj, Expr):
            symb_dict = {}
            expr_dict = {"name": str(obj), "expression": srepr(obj)}
            symbs: set[Parameter | Basic] = obj.free_symbols
            if symbs:
                symb_dict = {"symbols": {str(s): s._to_dict() for s in symbs}}
            d = {**expr_dict, **symb_dict}
        elif isinstance(obj, (QuantumModel, QNN, TransformedModule)):
            d = obj._to_dict(save_params)
        elif isinstance(obj, torch.nn.Module):
            d = {type(obj).__name__: obj.state_dict()}
        else:
            d = obj._to_dict()
    except Exception as e:
        logger.error(f"Serialization of object {obj} failed due to {e}")
    return d


def deserialize(d: dict, as_torch: bool = False) -> SUPPORTED_TYPES:
    """
    Supported Types:
    AbstractBlock | QuantumCircuit | QuantumModel | TransformedModule | Register | Module
    Deserializes a dict to one of the supported types.

    Arguments:
        d (dict): A dict containing a serialized object.
    Returns:
        AbstractBlock, QuantumCircuit, QuantumModel, TransformedModule, Register, Module.

    Examples:
    ```python exec="on" source="material-block" result="json"
    import torch
    from qadence import serialize, deserialize, hea, hamiltonian_factory, Z
    from qadence import QuantumCircuit, QuantumModel

    n_qubits = 2
    myblock = hea(n_qubits=n_qubits, depth=1)
    block_dict = serialize(myblock)
    print(block_dict)

    ## Lets use myblock in a QuantumCircuit and serialize it.

    qc = QuantumCircuit(n_qubits, myblock)
    qc_dict = serialize(qc)
    qc_deserialized = deserialize(qc_dict)
    assert qc == qc_deserialized

    ## Finally, let's wrap it in a QuantumModel
    obs = hamiltonian_factory(n_qubits, detuning = Z)
    qm = QuantumModel(qc, obs, backend='pyqtorch', diff_mode='ad')

    qm_dict = serialize(qm)
    qm_deserialized = deserialize(qm_dict)
    # Lets check if the loaded QuantumModel returns the same expectation
    assert torch.isclose(qm.expectation({}), qm_deserialized.expectation({}))
    ```
    """
    obj: Any
    if d.get("expression"):
        expr = eval(d["expression"])
        if hasattr(expr, "free_symbols"):
            for symb in expr.free_symbols:
                symb.value = float(d["symbols"][symb.name]["value"])
        obj = expr
    elif d.get("QuantumModel"):
        obj = QuantumModel._from_dict(d, as_torch)
    elif d.get("QNN"):
        obj = QNN._from_dict(d, as_torch)
    elif d.get("TransformedModule"):
        obj = TransformedModule._from_dict(d, as_torch)
    elif d.get("block") and d.get("register"):
        obj = QuantumCircuit._from_dict(d)
    elif d.get("graph"):
        obj = Register._from_dict(d)
    elif d.get("type"):
        if d["type"] in ALL_BLOCK_NAMES:
            block: AbstractBlock = (
                getattr(operations, d["type"])._from_dict(d)
                if hasattr(operations, d["type"])
                else getattr(qadenceblocks, d["type"])._from_dict(d)
            )
            if d["tag"] is not None:
                block = tag(block, d["tag"])
            obj = block
    else:
        import warnings

        msg = warnings.warn(
            "In order to load a custom torch.nn.Module, make sure its imported in the namespace."
        )
        try:
            module_name = list(d.keys())[0]
            obj = getattr(globals(), module_name)
            obj.load_state_dict(d[module_name])
        except Exception as e:
            logger.error(
                TypeError(
                    f"{msg}. Unable to deserialize object due to {e}.\
                    Supported objects are: {SUPPORTED_OBJECTS}"
                )
            )
    return obj


def save(
    obj: SUPPORTED_TYPES,
    folder: str | Path,
    file_name: str = "",
    format: SerializationFormat = SerializationFormat.JSON,
) -> None:
    """
    Same as serialize/deserialize but for storing/loading files.
    Supported types:
    AbstractBlock | QuantumCircuit | QuantumModel | TransformedModule | Register | torch.nn.Module
    Saves a qadence object to a json/.pt.

    Arguments:
        obj (AbstractBlock | QuantumCircuit | QuantumModel | Register):
                Either AbstractBlock, QuantumCircuit, QuantumModel, TransformedModule, Register.
        file_name (str): The name of the file.
        format (str): The type of file to save.
    Returns:
        None.

    Examples:
    ```python exec="on" source="material-block" result="json"
    import torch
    from pathlib import Path
    import os

    from qadence import save, load, hea, hamiltonian_factory, Z
    from qadence import QuantumCircuit, QuantumModel

    n_qubits = 2
    myblock = hea(n_qubits=n_qubits, depth=1)
    qc = QuantumCircuit(n_qubits, myblock)
    # Lets store the circuit in a json file
    save(qc, '.', 'circ')
    loaded_qc = load(Path('circ.json'))
    qc == loaded_qc
    os.remove('circ.json')
    ## Let's wrap it in a QuantumModel and store that
    obs = hamiltonian_factory(n_qubits, detuning = Z)
    qm = QuantumModel(qc, obs, backend='pyqtorch', diff_mode='ad')
    save(qm, folder= '.',file_name= 'quantum_model')
    qm_loaded = load('quantum_model.json')
    os.remove('quantum_model.json')
    ```
    """
    if not isinstance(obj, get_args(SUPPORTED_TYPES)):
        logger.error(f"Serialization of object type {type(obj)} not supported.")
    folder = Path(folder)
    if not folder.is_dir():
        logger.error(NotADirectoryError)
    if file_name == "":
        file_name = type(obj).__name__
    try:
        suffix, save_fn, _, save_params = FORMAT_DICT[format]
        d = serialize(obj, save_params)
        file_path = folder / Path(file_name + suffix)
        save_fn(d, file_path)
        logger.debug(f"Successfully saved {obj} from to {folder}.")
    except Exception as e:
        logger.error(f"Unable to write {type(obj)} to disk due to {e}")


def load(file_path: str | Path, map_location: str = "cpu") -> SUPPORTED_TYPES:
    """
    Same as serialize/deserialize but for storing/loading files.
    Supported types: AbstractBlock | QuantumCircuit | QuantumModel | TransformedModule | Register
    Loads a .json or .pt file to one of the supported types.

    Arguments:
        file_path (str): The name of the file.
        map_location (str): In case of a .pt file, on which device to load the object (cpu,cuda).
    Returns:
        A object of type AbstractBlock, QuantumCircuit, QuantumModel, TransformedModule, Register.

    Examples:
    ```python exec="on" source="material-block" result="json"
    import torch
    from pathlib import Path
    import os

    from qadence import save, load, hea, hamiltonian_factory, Z
    from qadence import QuantumCircuit, QuantumModel

    n_qubits = 2
    myblock = hea(n_qubits=n_qubits, depth=1)
    qc = QuantumCircuit(n_qubits, myblock)
    # Lets store the circuit in a json file
    save(qc, '.', 'circ')
    loaded_qc = load(Path('circ.json'))
    qc == loaded_qc
    os.remove('circ.json')
    ## Let's wrap it in a QuantumModel and store that
    obs = hamiltonian_factory(n_qubits, detuning = Z)
    qm = QuantumModel(qc, obs, backend='pyqtorch', diff_mode='ad')
    save(qm, folder= '.',file_name= 'quantum_model')
    qm_loaded = load('quantum_model.json')
    os.remove('quantum_model.json')
    ```
    """
    d = {}
    if isinstance(file_path, str):
        file_path = Path(file_path)
    if not os.path.exists(file_path):
        logger.error(f"File {file_path} not found.")
        raise FileNotFoundError
    FORMAT = file_extension(file_path)
    _, _, load_fn, _ = FORMAT_DICT[FORMAT]  # type: ignore[index]
    try:
        d = load_fn(file_path, map_location)
        logger.debug(f"Successfully loaded {d} from {file_path}.")
    except Exception as e:
        logger.error(f"Unable to load Object from {file_path} due to {e}")
    return deserialize(d)



from __future__ import annotations

import itertools
from typing import Any, Optional, Type, Union

from qadence.blocks import AbstractBlock, block_is_qubit_hamiltonian, chain, kron, tag
from qadence.operations import CNOT, CPHASE, CRX, CRY, CRZ, CZ, RX, RY, HamEvo
from qadence.types import Interaction, Strategy

from .hamiltonians import hamiltonian_factory
from .utils import build_idx_fms

DigitalEntanglers = Union[CNOT, CZ, CRZ, CRY, CRX]


def hea(
    n_qubits: int,
    depth: int = 1,
    param_prefix: str = "theta",
    support: tuple[int, ...] = None,
    strategy: Strategy = Strategy.DIGITAL,
    **strategy_args: Any,
) -> AbstractBlock:
    """
    Factory function for the Hardware Efficient Ansatz (HEA).

    Args:
        n_qubits: number of qubits in the block
        depth: number of layers of the HEA
        param_prefix: the base name of the variational parameters
        support: qubit indexes where the HEA is applied
        strategy: Strategy.Digital or Strategy.DigitalAnalog
        **strategy_args: see below

    Keyword Arguments:
        operations (list): list of operations to cycle through in the
            digital single-qubit rotations of each layer. Valid for
            Digital and DigitalAnalog HEA.
        periodic (bool): if the qubits should be linked periodically.
            periodic=False is not supported in emu-c. Valid for only
            for Digital HEA.
        entangler (AbstractBlock):
            - Digital: 2-qubit entangling operation. Supports CNOT, CZ,
            CRX, CRY, CRZ, CPHASE. Controlled rotations will have variational
            parameters on the rotation angles.
            - DigitaAnalog | Analog: Hamiltonian generator for the
            analog entangling layer. Defaults to global ZZ Hamiltonian.
            Time parameter is considered variational.


    Examples:
    ```python exec="on" source="material-block" result="json"
    from qadence import RZ, RX
    from qadence import hea

    # create the circuit
    n_qubits, depth = 2, 4
    ansatz = hea(
        n_qubits=n_qubits,
        depth=depth,
        strategy="sDAQC",
        operations=[RZ,RX,RZ]
    )
    ```
    """

    if support is None:
        support = tuple(range(n_qubits))

    hea_func_dict = {
        Strategy.DIGITAL: hea_digital,
        Strategy.SDAQC: hea_sDAQC,
        Strategy.BDAQC: hea_bDAQC,
        Strategy.ANALOG: hea_analog,
    }

    try:
        hea_func = hea_func_dict[strategy]
    except KeyError:
        raise KeyError(f"Strategy {strategy} not recognized.")

    hea_block: AbstractBlock = hea_func(
        n_qubits=n_qubits,
        depth=depth,
        param_prefix=param_prefix,
        support=support,
        **strategy_args,
    )  # type: ignore

    return hea_block


#############
## DIGITAL ##
#############


def _rotations_digital(
    n_qubits: int,
    depth: int,
    param_prefix: str = "theta",
    support: tuple[int, ...] = None,
    operations: list[Type[AbstractBlock]] = [RX, RY, RX],
) -> list[AbstractBlock]:
    """
    Creates the layers of single qubit rotations in an HEA.
    """
    if support is None:
        support = tuple(range(n_qubits))
    iterator = itertools.count()
    rot_list: list[AbstractBlock] = []
    for d in range(depth):
        rots = [
            kron(
                gate(support[n], param_prefix + f"_{next(iterator)}")  # type: ignore [arg-type]
                for n in range(n_qubits)
            )
            for gate in operations
        ]
        rot_list.append(chain(*rots))
    return rot_list


def _entangler(
    control: int,
    target: int,
    param_str: str,
    op: Type[DigitalEntanglers] = CNOT,
) -> AbstractBlock:
    if op in [CNOT, CZ]:
        return op(control, target)  # type: ignore
    elif op in [CRZ, CRY, CRX, CPHASE]:
        return op(control, target, param_str)  # type: ignore
    else:
        raise ValueError("Provided entangler not accepted for digital HEA.")


def _entanglers_digital(
    n_qubits: int,
    depth: int,
    param_prefix: str = "theta",
    support: tuple[int, ...] = None,
    periodic: bool = False,
    entangler: Type[DigitalEntanglers] = CNOT,
) -> list[AbstractBlock]:
    """
    Creates the layers of digital entangling operations in an HEA.
    """
    if support is None:
        support = tuple(range(n_qubits))
    iterator = itertools.count()
    ent_list: list[AbstractBlock] = []
    for d in range(depth):
        ents = []
        ents.append(
            kron(
                _entangler(
                    control=support[n],
                    target=support[n + 1],
                    param_str=param_prefix + f"_ent_{next(iterator)}",
                    op=entangler,
                )
                for n in range(n_qubits)
                if not n % 2 and n < n_qubits - 1
            )
        )
        if n_qubits > 2:
            ents.append(
                kron(
                    _entangler(
                        control=support[n],
                        target=support[(n + 1) % n_qubits],
                        param_str=param_prefix + f"_ent_{next(iterator)}",
                        op=entangler,
                    )
                    for n in range(n_qubits - (not periodic))
                    if n % 2
                )
            )
        ent_list.append(chain(*ents))
    return ent_list


def hea_digital(
    n_qubits: int,
    depth: int = 1,
    param_prefix: str = "theta",
    periodic: bool = False,
    operations: list[type[AbstractBlock]] = [RX, RY, RX],
    support: tuple[int, ...] = None,
    entangler: Type[DigitalEntanglers] = CNOT,
) -> AbstractBlock:
    """
    Construct the Digital Hardware Efficient Ansatz (HEA).

    Args:
        n_qubits (int): number of qubits in the block.
        depth (int): number of layers of the HEA.
        param_prefix (str): the base name of the variational parameters
        periodic (bool): if the qubits should be linked periodically.
            periodic=False is not supported in emu-c.
        operations (list): list of operations to cycle through in the
            digital single-qubit rotations of each layer.
        support (tuple): qubit indexes where the HEA is applied.
        entangler (AbstractBlock): 2-qubit entangling operation.
            Supports CNOT, CZ, CRX, CRY, CRZ. Controlld rotations
            will have variational parameters on the rotation angles.
    """
    try:
        if entangler not in [CNOT, CZ, CRX, CRY, CRZ, CPHASE]:
            raise ValueError(
                "Please provide a valid two-qubit entangler operation for digital HEA."
            )
    except TypeError:
        raise ValueError("Please provide a valid two-qubit entangler operation for digital HEA.")

    rot_list = _rotations_digital(
        n_qubits=n_qubits,
        depth=depth,
        param_prefix=param_prefix,
        support=support,
        operations=operations,
    )

    ent_list = _entanglers_digital(
        n_qubits=n_qubits,
        depth=depth,
        param_prefix=param_prefix,
        support=support,
        periodic=periodic,
        entangler=entangler,
    )

    layers = []
    for d in range(depth):
        layers.append(rot_list[d])
        layers.append(ent_list[d])
    return tag(chain(*layers), "HEA")


###########
## sDAQC ##
###########


def _entanglers_analog(
    depth: int,
    param_prefix: str = "theta",
    entangler: AbstractBlock | None = None,
) -> list[AbstractBlock]:
    return [HamEvo(entangler, param_prefix + f"_t_{d}") for d in range(depth)]


def hea_sDAQC(
    n_qubits: int,
    depth: int = 1,
    param_prefix: str = "theta",
    operations: list[type[AbstractBlock]] = [RX, RY, RX],
    support: tuple[int, ...] = None,
    entangler: AbstractBlock | None = None,
) -> AbstractBlock:
    """
    Construct the Hardware Efficient Ansatz (HEA) with analog entangling layers
    using step-wise digital-analog computation.

    Args:
        n_qubits (int): number of qubits in the block.
        depth (int): number of layers of the HEA.
        param_prefix (str): the base name of the variational parameters
        operations (list): list of operations to cycle through in the
            digital single-qubit rotations of each layer.
        support (tuple): qubit indexes where the HEA is applied.
        entangler (AbstractBlock): Hamiltonian generator for the
            analog entangling layer. Defaults to global ZZ Hamiltonian.
            Time parameter is considered variational.
    """

    # TODO: Add qubit support
    if entangler is None:
        entangler = hamiltonian_factory(n_qubits, interaction=Interaction.NN)
    try:
        if not block_is_qubit_hamiltonian(entangler):
            raise ValueError(
                "Please provide a valid Pauli Hamiltonian generator for digital-analog HEA."
            )
    except NotImplementedError:
        raise ValueError(
            "Please provide a valid Pauli Hamiltonian generator for digital-analog HEA."
        )

    rot_list = _rotations_digital(
        n_qubits=n_qubits,
        depth=depth,
        param_prefix=param_prefix,
        support=support,
        operations=operations,
    )

    ent_list = _entanglers_analog(
        depth=depth,
        param_prefix=param_prefix,
        entangler=entangler,
    )

    layers = []
    for d in range(depth):
        layers.append(rot_list[d])
        layers.append(ent_list[d])
    return tag(chain(*layers), "HEA-sDA")


###########
## bDAQC ##
###########


def hea_bDAQC(*args: Any, **kwargs: Any) -> Any:
    raise NotImplementedError


############
## ANALOG ##
############


def hea_analog(*args: Any, **kwargs: Any) -> Any:
    raise NotImplementedError


#########
## QNN ##
#########


def build_qnn(
    n_qubits: int,
    n_features: int,
    depth: int = None,
    ansatz: Optional[AbstractBlock] = None,
    fm_pauli: Type[RY] = RY,
    spectrum: str = "simple",
    basis: str = "fourier",
    fm_strategy: str = "parallel",
) -> list[AbstractBlock]:
    """Helper function to build a qadence QNN quantum circuit

    Args:
        n_qubits (int): The number of qubits.
        n_features (int): The number of input dimensions.
        depth (int): The depth of the ansatz.
        ansatz (Optional[AbstractBlock]):  An optional argument to pass a custom qadence ansatz.
        fm_pauli (str): The type of Pauli gate for the feature map. Must be one of 'RX',
            'RY', or 'RZ'.
        spectrum (str): The desired spectrum of the feature map generator. The options simple,
            tower and exponential produce a spectrum with linear, quadratic and exponential
            eigenvalues with respect to the number of qubits.
        basis (str): The encoding function. The options fourier and chebyshev correspond to Φ(x)=x
            and arcos(x) respectively.
        fm_strategy (str): The feature map encoding strategy. If "parallel", the features
            are encoded in one block of rotation gates, with each feature given
            an equal number of qubits. If "serial", the features are encoded
            sequentially, with a HEA block between.

    Returns:
        A list of Abstract blocks to be used for constructing a quantum circuit
    """
    depth = n_qubits if depth is None else depth

    idx_fms = build_idx_fms(basis, fm_pauli, fm_strategy, n_features, n_qubits, spectrum)

    if fm_strategy == "parallel":
        _fm = kron(*idx_fms)
        fm = tag(_fm, tag="FM")

    elif fm_strategy == "serial":
        fm_components: list[AbstractBlock] = []
        for j, fm_idx in enumerate(idx_fms[:-1]):
            fm_idx = tag(fm_idx, tag=f"FM{j}")  # type: ignore[assignment]
            fm_component = (fm_idx, hea(n_qubits, 1, f"theta_{j}"))
            fm_components.extend(fm_component)
        fm_components.append(tag(idx_fms[-1], tag=f"FM{len(idx_fms) - 1}"))
        fm = chain(*fm_components)  # type: ignore[assignment]

    ansatz = hea(n_qubits, depth=depth) if ansatz is None else ansatz
    return [fm, ansatz]



from __future__ import annotations

from typing import Callable

import nevergrad as ng
from nevergrad.optimization.base import Optimizer as NGOptimizer
from rich.progress import BarColumn, Progress, TaskProgressColumn, TextColumn, TimeRemainingColumn
from torch import Tensor
from torch.nn import Module
from torch.utils.data import DataLoader
from torch.utils.tensorboard import SummaryWriter

from qadence.logger import get_logger
from qadence.ml_tools.config import TrainConfig
from qadence.ml_tools.data import DictDataLoader
from qadence.ml_tools.parameters import get_parameters, set_parameters
from qadence.ml_tools.printing import print_metrics, write_tensorboard
from qadence.ml_tools.saveload import load_checkpoint, write_checkpoint
from qadence.ml_tools.tensors import promote_to_tensor

logger = get_logger(__name__)


def train(
    model: Module,
    dataloader: DictDataLoader | DataLoader | None,
    optimizer: NGOptimizer,
    config: TrainConfig,
    loss_fn: Callable,
) -> tuple[Module, NGOptimizer]:
    """Runs the training loop with a gradient-free optimizer

    Assumes that `loss_fn` returns a tuple of (loss, metrics: dict), where
    `metrics` is a dict of scalars. Loss and metrics are written to
    tensorboard. Checkpoints are written every `config.checkpoint_every` steps
    (and after the last training step).  If a checkpoint is found at `config.folder`
    we resume training from there.  The tensorboard logs can be viewed via
    `tensorboard --logdir /path/to/folder`.

    Args:
        model: The model to train
        dataloader: Dataloader constructed via `dictdataloader`
        optimizer: The optimizer to use taken from the Nevergrad library. If this is not
            the case the function will raise an AssertionError
        loss_fn: Loss function returning (loss: float, metrics: dict[str, float])
    """
    init_iter = 0
    if config.folder:
        model, optimizer, init_iter = load_checkpoint(config.folder, model, optimizer)
        logger.debug(f"Loaded model and optimizer from {config.folder}")

    def _update_parameters(
        data: Tensor | None, ng_params: ng.p.Array
    ) -> tuple[float, dict, ng.p.Array]:
        loss, metrics = loss_fn(model, data)  # type: ignore[misc]
        optimizer.tell(ng_params, float(loss))
        ng_params = optimizer.ask()  # type: ignore [assignment]
        params = promote_to_tensor(ng_params.value, requires_grad=False)
        set_parameters(model, params)
        return loss, metrics, ng_params

    assert loss_fn is not None, "Provide a valid loss function"
    # TODO: support also Scipy optimizers
    assert isinstance(optimizer, NGOptimizer), "Use only optimizers from the Nevergrad library"

    # initialize tensorboard
    writer = SummaryWriter(config.folder, purge_step=init_iter)

    # set optimizer configuration and initial parameters
    optimizer.budget = config.max_iter
    optimizer.enable_pickling()

    # TODO: Make it GPU compatible if possible
    params = get_parameters(model).detach().numpy()
    ng_params = ng.p.Array(init=params)

    # serial training
    # TODO: Add a parallelization using the num_workers argument in Nevergrad
    progress = Progress(
        TextColumn("[progress.description]{task.description}"),
        BarColumn(),
        TaskProgressColumn(),
        TimeRemainingColumn(elapsed_when_finished=True),
    )
    with progress:
        dl_iter = iter(dataloader) if isinstance(dataloader, DictDataLoader) else None

        for iteration in progress.track(range(init_iter, init_iter + config.max_iter)):
            if dataloader is None:
                loss, metrics, ng_params = _update_parameters(None, ng_params)

            elif isinstance(dataloader, DictDataLoader):
                # resample all the time from the dataloader
                # by creating a fresh iterator if the dataloader
                # does not support automatically iterating datasets
                if not dataloader.has_automatic_iter:
                    dl_iter = iter(dataloader)

                data = next(dl_iter)  # type: ignore[arg-type]
                loss, metrics, ng_params = _update_parameters(data, ng_params)

            elif isinstance(dataloader, DataLoader):
                # single-epoch with standard DataLoader
                # otherwise a standard PyTorch DataLoader behavior
                # is assumed with optional mini-batches
                running_loss = 0.0
                for i, data in enumerate(dataloader):
                    loss, metrics, ng_params = _update_parameters(data, ng_params)
                    running_loss += loss
                loss = running_loss / (i + 1)

            else:
                raise NotImplementedError("Unsupported dataloader type!")

            if iteration % config.print_every == 0:
                print_metrics(loss, metrics, iteration)

            if iteration % config.write_every == 0:
                write_tensorboard(writer, loss, metrics, iteration)

            if config.folder:
                if iteration % config.checkpoint_every == 0:
                    write_checkpoint(config.folder, model, optimizer, iteration)

            if iteration >= init_iter + config.max_iter:
                break

    ## Final writing and stuff
    if config.folder:
        write_checkpoint(config.folder, model, optimizer, iteration)
    write_tensorboard(writer, loss, metrics, iteration)
    writer.close()

    return model, optimizer



from __future__ import annotations

import io
from typing import Any

from graphviz import Graph

from .themes import Dark, Light
from .utils import make_diagram
from .vizbackend import QuantumCircuitDiagram


def display(x: Any, *args: Any, **kwargs: Any) -> Graph:
    return make_diagram(x, *args, **kwargs).show()


def savefig(x: Any, filename: str, *args: Any, **kwargs: Any) -> None:
    make_diagram(x, *args, **kwargs).savefig(filename)


def html_string(x: Any, *args: Any, **kwargs: Any) -> str:
    buffer = io.StringIO()

    qcd = make_diagram(x, *args, **kwargs)
    qcd._build()

    buffer.write(qcd.graph.pipe(format="svg").decode("utf-8"))
    buffer.seek(0)
    return buffer.read()



from __future__ import annotations

from uuid import UUID

import torch

from qadence.blocks import (
    AbstractBlock,
    AddBlock,
    ChainBlock,
    ControlBlock,
    KronBlock,
    ParametricBlock,
    ParametricControlBlock,
    PrimitiveBlock,
    ScaleBlock,
)
from qadence.blocks.utils import chain, kron, uuid_to_expression
from qadence.parameters import evaluate, stringify
from qadence.types import Endianness, TensorType, TNumber

J = torch.tensor(1j)

ZEROMAT = torch.zeros((2, 2), dtype=torch.cdouble).unsqueeze(0)
IMAT = torch.eye(2, dtype=torch.cdouble).unsqueeze(0)
XMAT = torch.tensor([[0, 1], [1, 0]], dtype=torch.cdouble).unsqueeze(0)
YMAT = torch.tensor([[0, -1j], [1j, 0]], dtype=torch.cdouble).unsqueeze(0)
ZMAT = torch.tensor([[1, 0], [0, -1]], dtype=torch.cdouble).unsqueeze(0)
NMAT = torch.tensor([[0, 0], [0, 1]], dtype=torch.cdouble).unsqueeze(0)
SMAT = torch.tensor([[1, 0], [0, 1j]], dtype=torch.cdouble).unsqueeze(0)
SDAGMAT = torch.tensor([[1, 0], [0, -1j]], dtype=torch.cdouble).unsqueeze(0)
TMAT = torch.tensor([[1, 0], [0, torch.exp(J * torch.pi / 4)]], dtype=torch.cdouble).unsqueeze(0)
TDAGMAT = torch.tensor([[1, 0], [0, torch.exp(J * torch.pi / 4)]], dtype=torch.cdouble).unsqueeze(0)
HMAT = (
    1
    / torch.sqrt(torch.tensor(2))
    * torch.tensor([[1, 1], [1, -1]], dtype=torch.cdouble).unsqueeze(0)
)


OPERATIONS_DICT = {
    "Zero": ZEROMAT,
    "I": IMAT,
    "X": XMAT,
    "Y": YMAT,
    "Z": ZMAT,
    "S": SMAT,
    "SDagger": SDAGMAT,
    "T": TMAT,
    "TDagger": TDAGMAT,
    "H": HMAT,
    "N": NMAT,
}


def _fill_identities(
    block_mat: torch.Tensor,
    qubit_support: tuple,
    full_qubit_support: tuple | list,
    diag_only: bool = False,
    endianness: Endianness = Endianness.BIG,
) -> torch.Tensor:
    """Returns a Kronecker product of matrix defined on a subset of qubits with identities
    acting on the unused qubits.

    Args:
        block_mat (torch.Tensor): matrix of an arbitrary gate
        qubit_support (tuple): qubit support of `block_mat`
        full_qubit_support (tuple): full qubit support of the circuit
        diag_only (bool): Use diagonals only

    Returns:
        torch.Tensor: augmented matrix with dimensions (2**nqubits, 2**nqubits)
        or a tensor (2**n_qubits) if diag_only
    """
    qubit_support = tuple(sorted(qubit_support))
    mat = IMAT if qubit_support[0] != full_qubit_support[0] else block_mat
    if diag_only:
        mat = torch.diag(mat.squeeze(0))
    for i in full_qubit_support[1:]:
        if i == qubit_support[0]:
            other = torch.diag(block_mat.squeeze(0)) if diag_only else block_mat
            if endianness == Endianness.LITTLE:
                mat = torch.kron(other, mat)
            else:
                mat = torch.kron(mat, other)
        elif i not in qubit_support:
            other = torch.diag(IMAT.squeeze(0)) if diag_only else IMAT
            if endianness == Endianness.LITTLE:
                mat = torch.kron(other.contiguous(), mat.contiguous())
            else:
                mat = torch.kron(mat.contiguous(), other.contiguous())

    return mat


def _rot_matrices(theta: torch.Tensor, generator: torch.Tensor) -> torch.Tensor:
    """
    Args:
        theta(torch.Tensor): input parameter
        generator(torch.Tensor): the tensor of the generator
    Returns:
        torch.Tensor: a batch of gates after applying theta
    """
    batch_size = theta.size(0)

    cos_t = torch.cos(theta / 2).unsqueeze(1).unsqueeze(2)
    cos_t = cos_t.repeat((1, 2, 2))
    sin_t = torch.sin(theta / 2).unsqueeze(1).unsqueeze(2)
    sin_t = sin_t.repeat((1, 2, 2))

    batch_imat = IMAT.repeat(batch_size, 1, 1)
    batch_generator = generator.repeat(batch_size, 1, 1)

    return cos_t * batch_imat - 1j * sin_t * batch_generator


def _u_matrix(theta: tuple[torch.Tensor, ...]) -> torch.Tensor:
    """
    Args:
        theta(tuple[torch.Tensor]): tuple of torch Tensor with 3 elements
            per each parameter of the arbitrary rotation
    Returns:
        torch.Tensor: matrix corresponding to the U gate after applying theta
    """
    z_phi = _rot_matrices(theta[0], OPERATIONS_DICT["Z"])
    y_theta = _rot_matrices(theta[1], OPERATIONS_DICT["Y"])
    z_omega = _rot_matrices(theta[2], OPERATIONS_DICT["Z"])

    res = torch.matmul(y_theta, z_phi)
    res = torch.matmul(z_omega, res)
    return res


def _phase_matrix(theta: torch.Tensor | TNumber) -> torch.Tensor:
    """
    Args:
        theta(torch.Tensor): input parameter
    Returns:
        torch.Tensor: a batch of gates after applying theta
    """
    exp_t = torch.exp(1j * theta).unsqueeze(1).unsqueeze(2)
    exp_t = exp_t.repeat((1, 2, 2))
    return 0.5 * (IMAT + ZMAT) + exp_t * 0.5 * (IMAT - ZMAT)


def _parametric_matrix(gate: ParametricBlock, values: dict[str, torch.Tensor]) -> torch.Tensor:
    from qadence.operations import PHASE, RX, RY, RZ, U

    theta = _gate_parameters(gate, values)
    if isinstance(gate, (RX, RY, RZ)):
        pmat = _rot_matrices(
            theta[0], OPERATIONS_DICT[gate.generator.name]  # type:ignore[union-attr]
        )
    elif isinstance(gate, U):
        pmat = _u_matrix(theta)
    elif isinstance(gate, PHASE):
        pmat = _phase_matrix(theta[0])
    return pmat


def _controlled_block_with_params(
    block: ParametricControlBlock | ControlBlock,
) -> tuple[AbstractBlock, dict[str, torch.Tensor]]:
    """Redefines parameterized/non-parameterized controlled block.

    Args:
        block (ParametricControlBlock): original controlled rotation block

    Returns:
        AbstractBlock: redefined controlled rotation block
        dict with new parameters which are added
    """
    from qadence.operations import I, Z

    # redefine controlled rotation block in a way suitable for matrix evaluation
    control = block.qubit_support[:-1]
    target = block.qubit_support[-1]
    p1 = kron(0.5 * I(qubit) + (-0.5) * Z(qubit) for qubit in control)
    p0 = I(control[0]) - p1
    c_block = kron(p0, I(target)) + kron(p1, block.blocks[0])

    uuid_expr = uuid_to_expression(c_block)
    newparams = {
        stringify(expr): evaluate(expr, {}, as_torch=True)
        for uuid, expr in uuid_expr.items()
        if expr.is_number
    }

    return c_block, newparams


def _swap_block(block: AbstractBlock) -> AbstractBlock:
    """Redefines SWAP block.

    Args:
        block (AbstractBlock): original SWAP block

    Returns:
        AbstractBlock: redefined SWAP block
    """
    from qadence.operations import CNOT

    # redefine controlled rotation block in a way suitable for matrix evaluation
    control = block.qubit_support[0]
    target = block.qubit_support[1]
    swap_block = chain(CNOT(control, target), CNOT(target, control), CNOT(control, target))

    return swap_block


def _cswap_block(block: AbstractBlock) -> torch.Tensor:
    from qadence.operations import Toffoli

    control = block.qubit_support[0]
    target1 = block.qubit_support[1]
    target2 = block.qubit_support[2]

    cswap_block = chain(
        Toffoli((control, target2), target1),
        Toffoli((control, target1), target2),
        Toffoli((control, target2), target1),
    )

    return cswap_block


def _extract_param_names_or_uuids(b: AbstractBlock, uuids: bool = False) -> tuple[str, ...]:
    if isinstance(b, ParametricBlock):
        return (
            tuple(b.parameters.uuids())
            if uuids
            else tuple(map(stringify, b.parameters.expressions()))
        )
    else:
        return ()


def is_valid_uuid(value: str) -> bool:
    try:
        UUID(value)
        return True
    except ValueError:
        return False


def _gate_parameters(b: AbstractBlock, values: dict[str, torch.Tensor]) -> tuple[torch.Tensor, ...]:
    uuids = is_valid_uuid(list(values.keys())[0])
    ks = _extract_param_names_or_uuids(b, uuids=uuids)
    return tuple(values[k] for k in ks)


def block_to_diagonal(
    block: AbstractBlock,
    qubit_support: tuple | list | None = None,
    use_full_support: bool = True,
    endianness: Endianness = Endianness.BIG,
) -> torch.Tensor:
    if block.is_parametric:
        raise TypeError("Sparse observables cant be parametric.")
    if not block._is_diag_pauli:
        raise TypeError("Sparse observables can only be used on paulis which are diagonal.")
    if qubit_support is None:
        if use_full_support:
            qubit_support = tuple(range(0, block.n_qubits))
        else:
            qubit_support = block.qubit_support
    nqubits = len(qubit_support)  # type: ignore [arg-type]
    if isinstance(block, (ChainBlock, KronBlock)):
        v = torch.ones(2**nqubits, dtype=torch.cdouble)
        for b in block.blocks:
            v *= block_to_diagonal(b, qubit_support)
    if isinstance(block, AddBlock):
        t = torch.zeros(2**nqubits, dtype=torch.cdouble)
        for b in block.blocks:
            t += block_to_diagonal(b, qubit_support)
        v = t
    elif isinstance(block, ScaleBlock):
        _s = evaluate(block.scale, {}, as_torch=True)  # type: ignore[attr-defined]
        _s = _s.detach()  # type: ignore[union-attr]
        v = _s * block_to_diagonal(block.block, qubit_support)

    elif isinstance(block, PrimitiveBlock):
        v = _fill_identities(
            OPERATIONS_DICT[block.name],
            block.qubit_support,
            qubit_support,  # type: ignore [arg-type]
            diag_only=True,
            endianness=endianness,
        )
    return v


# version that will accept user params
def block_to_tensor(
    block: AbstractBlock,
    values: dict[str, TNumber | torch.Tensor] = {},
    qubit_support: tuple | None = None,
    use_full_support: bool = True,
    tensor_type: TensorType = TensorType.DENSE,
    endianness: Endianness = Endianness.BIG,
) -> torch.Tensor:
    """
    Convert a block into a torch tensor.

    Arguments:
        block (AbstractBlock): The block to convert.
        values (dict): A optional dict with values for parameters.
        qubit_support (tuple): The qubit_support of the block.
        use_full_support (bool): True infers the total number of qubits.
        tensor_type (TensorType): the target tensor type.

    Returns:
        A torch.Tensor.

    Examples:
    ```python exec="on" source="material-block" result="json"
    from qadence import hea, hamiltonian_factory, Z, block_to_tensor

    block = hea(2,2)
    print(block_to_tensor(block))

    # In case you have a diagonal observable, you can use
    obs = hamiltonian_factory(2, detuning = Z)
    print(block_to_tensor(obs, tensor_type="SparseDiagonal"))
    ```
    """

    # FIXME: default use_full_support to False. In general, it would
    # be more efficient to do that, and make sure that computations such
    # as observables only do the matmul of the size of the qubit support.

    if tensor_type == TensorType.DENSE:
        from qadence.blocks import embedding

        (ps, embed) = embedding(block)
        return _block_to_tensor_embedded(
            block, embed(ps, values), qubit_support, use_full_support, endianness=endianness
        )

    elif tensor_type == TensorType.SPARSEDIAGONAL:
        t = block_to_diagonal(block, endianness=endianness)
        indices, values, size = torch.nonzero(t), t[t != 0], len(t)
        indices = torch.stack((indices.flatten(), indices.flatten()))
        return torch.sparse_coo_tensor(indices, values, (size, size))


# version that accepts embedded params
def _block_to_tensor_embedded(
    block: AbstractBlock,
    values: dict[str, TNumber | torch.Tensor] = {},
    qubit_support: tuple | None = None,
    use_full_support: bool = True,
    endianness: Endianness = Endianness.BIG,
) -> torch.Tensor:
    from qadence.blocks import MatrixBlock
    from qadence.operations import CSWAP, SWAP, HamEvo

    # get number of qubits
    if qubit_support is None:
        if use_full_support:
            qubit_support = tuple(range(0, block.n_qubits))
        else:
            qubit_support = block.qubit_support
    nqubits = len(qubit_support)

    if isinstance(block, (ChainBlock, KronBlock)):
        # create identity matrix of appropriate dimensions
        mat = IMAT.clone()
        for i in range(nqubits - 1):
            mat = torch.kron(mat, IMAT)

        # perform matrix multiplications
        for b in block.blocks:
            other = _block_to_tensor_embedded(b, values, qubit_support, endianness=endianness)
            mat = torch.matmul(other, mat)

    elif isinstance(block, AddBlock):
        # create zero matrix of appropriate dimensions
        mat = ZEROMAT.clone()
        for _ in range(nqubits - 1):
            mat = torch.kron(mat, ZEROMAT)

        # perform matrix summation
        for b in block.blocks:
            mat = mat + _block_to_tensor_embedded(b, values, qubit_support, endianness=endianness)

    elif isinstance(block, HamEvo):
        if block.qubit_support:
            if isinstance(block.generator, AbstractBlock):
                # get matrix representation of generator
                gen_mat = _block_to_tensor_embedded(
                    block.generator, values, qubit_support, endianness=endianness
                )

                # calculate evolution matrix
                (p,) = _gate_parameters(block, values)
                prefac = -J * p
                mat = torch.linalg.matrix_exp(prefac * gen_mat)
            elif isinstance(block.generator, torch.Tensor):
                gen_mat = block.generator

                # calculate evolution matrix
                (p, _) = _gate_parameters(block, values)
                prefac = -J * p
                mat = torch.linalg.matrix_exp(prefac * gen_mat)

                # add missing identities on unused qubits
                mat = _fill_identities(
                    mat, block.qubit_support, qubit_support, endianness=endianness
                )
            else:
                raise TypeError(
                    f"Generator of type {type(block.generator)} not supported in HamEvo."
                )
        else:
            raise ValueError("qubit_support is not defined for HamEvo block.")

        mat = mat.unsqueeze(0) if len(mat.size()) == 2 else mat

    elif isinstance(block, CSWAP):
        cswap_block = _cswap_block(block)
        mat = _block_to_tensor_embedded(cswap_block, values, qubit_support, endianness=endianness)

    elif isinstance(block, (ControlBlock, ParametricControlBlock)):
        c_block, newparams = _controlled_block_with_params(block)
        newparams.update(values)
        mat = _block_to_tensor_embedded(c_block, newparams, qubit_support, endianness=endianness)

    elif isinstance(block, ScaleBlock):
        (scale,) = _gate_parameters(block, values)
        mat = scale * _block_to_tensor_embedded(
            block.block, values, qubit_support, endianness=endianness
        )

    elif isinstance(block, ParametricBlock):
        block_mat = _parametric_matrix(block, values)

        # add missing identities on unused qubits
        mat = _fill_identities(block_mat, block.qubit_support, qubit_support, endianness=endianness)

    elif isinstance(block, MatrixBlock):
        mat = block.matrix.unsqueeze(0)
        # FIXME: properly handle identity filling in matrix blocks
        # mat = _fill_identities(
        #    block.matrix.unsqueeze(0),
        #    block.qubit_support,
        #    qubit_support,
        #    endianness=endianness,
        # )

    elif isinstance(block, SWAP):
        swap_block = _swap_block(block)
        mat = _block_to_tensor_embedded(swap_block, values, qubit_support, endianness=endianness)

    elif block.name in OPERATIONS_DICT.keys():
        block_mat = OPERATIONS_DICT[block.name]

        # add missing identities on unused qubits
        mat = _fill_identities(block_mat, block.qubit_support, qubit_support, endianness=endianness)

    else:
        raise TypeError(f"Conversion for block type {type(block)} not supported.")

    return mat



from __future__ import annotations

import datetime
import os
from dataclasses import dataclass
from pathlib import Path
from typing import Callable, Optional


@dataclass
class TrainConfig:
    """Default config for the train function. The default value of
    each field can be customize with the constructor:

    ```python exec="on" source="material-block" result="json"
    from qadence.ml_tools import TrainConfig
    c = TrainConfig(folder="/tmp/train")
    print(str(c)) # markdown-exec: hide
    ```
    """

    max_iter: int = 10000
    """Number of training iterations."""
    print_every: int = 1000
    """Print loss/metrics."""
    write_every: int = 50
    """Write tensorboard logs"""
    checkpoint_every: int = 5000
    """Write model/optimizer checkpoint"""
    folder: Optional[Path] = None
    """Checkpoint/tensorboard logs folder"""
    create_subfolder_per_run: bool = False
    """Checkpoint/tensorboard logs stored in subfolder with name `<timestamp>_<PID>`.
    Prevents continuing from previous checkpoint, useful for fast prototyping."""
    checkpoint_best_only: bool = False
    """Write model/optimizer checkpoint only if a metric has improved"""
    validation_criterion: Optional[Callable] = None
    """A boolean function which evaluates a given validation metric is satisfied"""
    trainstop_criterion: Optional[Callable] = None
    """A boolean function which evaluates a given training stopping metric is satisfied"""
    batch_size: int = 1
    """The batch_size to use when passing a list/tuple of torch.Tensors."""

    def __post_init__(self) -> None:
        if self.folder:
            if isinstance(self.folder, str):  # type: ignore [unreachable]
                self.folder = Path(self.folder)  # type: ignore [unreachable]
            if self.create_subfolder_per_run:
                subfoldername = (
                    datetime.datetime.now().strftime("%Y%m%dT%H%M%S") + "_" + hex(os.getpid())[2:]
                )
                self.folder = self.folder / subfoldername
        if self.trainstop_criterion is None:
            self.trainstop_criterion = lambda x: x <= self.max_iter
        if self.validation_criterion is None:
            self.validation_criterion = lambda x: False



from __future__ import annotations

import torch

from qadence.blocks import AbstractBlock, KronBlock
from qadence.blocks.utils import unroll_block_with_scaling
from qadence.logger import get_logger
from qadence.operations import N, Z
from qadence.parameters import Parameter, evaluate
from qadence.types import GenDAQC

from .utils import _ix_map

logger = get_logger(__name__)


def _parse_generator(
    n_qubits: int,
    generator: AbstractBlock,
    zero_tol: float,
) -> torch.Tensor:
    """
    Parses the input generator to extract the `g_jk` weights
    of the Ising model and the respective target qubits `(j, k)`.
    """

    flat_size = int(0.5 * n_qubits * (n_qubits - 1))
    g_jk_list = torch.zeros(flat_size)
    g_jk_mat = torch.zeros(n_qubits, n_qubits)

    # This parser is heavily dependent on unroll_block_with_scaling
    gen_list = unroll_block_with_scaling(generator)

    # Now we wish to check if generator is of the form:
    # `add(g_jk * kron(op(j), op(k)) for j < k)`
    # and determine if `op = Z` or `op = N`

    gen_type_Z = []
    gen_type_N = []

    for block, scale in gen_list:
        if isinstance(scale, Parameter):
            raise TypeError("DAQC transform does not support parameterized Hamiltonians.")

        # First we check if all relevant blocks (with non-negligible scaling)
        # are of type(KronBlock), since we only admit kron(Z, Z) or kron(N, N).
        if not isinstance(block, KronBlock):
            if abs(scale) < zero_tol:
                continue
            else:
                raise TypeError(
                    "DAQC transform only supports ZZ or NN interaction Hamiltonians."
                    "Error found on block: {block}."
                )

        # Next we check and keep track of the contents of each KronBlock
        for pauli in block.blocks:
            if isinstance(pauli, Z):
                gen_type_Z.append(True)
                gen_type_N.append(False)
            elif isinstance(pauli, N):
                gen_type_N.append(True)
                gen_type_Z.append(False)
            else:
                raise ValueError(
                    "DAQC transform only supports ZZ or NN interaction Hamiltonians."
                    "Error found on block: {block}."
                )

        # We save the qubit support and interaction
        # strength of each KronBlock to be used in DAQC
        j, k = block.qubit_support
        g_jk = torch.tensor(evaluate(scale), dtype=torch.get_default_dtype())

        beta = _ix_map(n_qubits, j + 1, k + 1)

        # Flat list of interaction strength
        g_jk_list[beta - 1] += g_jk

        # Symmetric matrix of interaction strength
        g_jk_mat[j, k] += g_jk
        g_jk_mat[k, j] += g_jk

    # Finally we check if all individual interaction terms were
    # either ZZ or NN to determine the generator type.
    if torch.tensor(gen_type_Z).prod() == 1 and len(gen_type_Z) > 0:
        gen_type = GenDAQC.ZZ
    elif torch.tensor(gen_type_N).prod() == 1 and len(gen_type_N) > 0:
        gen_type = GenDAQC.NN
    else:
        raise ValueError(
            "Wrong Hamiltonian structure provided. "
            "Possible mixture of Z and N terms in the Hamiltonian."
        )

    g_jk_list[g_jk_list == 0.0] = zero_tol

    return g_jk_list, g_jk_mat, gen_type


def _check_compatibility(
    g_jk_target: torch.Tensor,
    g_jk_build: torch.Tensor,
    zero_tol: float,
) -> bool:
    """
    Checks if the build Hamiltonian is missing any interactions needed
    for the transformation into the requested target Hamiltonian.
    """
    for g_t, g_b in zip(g_jk_target, g_jk_build):
        if abs(g_t) > zero_tol and abs(g_b) <= zero_tol:
            raise ValueError("Incompatible interactions between target and build Hamiltonians.")
    return True
